@@ -1,9 +1,7 @@
 #! /bin/false
 
-# $Id: Walk.pm,v 1.15 2006/05/11 14:10:54 guido Exp $
-
 # Traverse Perl data structures.
-# Copyright (C) 2005-2006 Guido Flohr <guido@imperia.net>,
+# Copyright (C) 2005-2016 Guido Flohr <guido.flohr@cantanea.com>,
 # all rights reserved.
 
 # This program is free software; you can redistribute it and/or modify it
@@ -26,15 +24,17 @@ package Data::Walk;
 use strict;
 use 5.004;
 
+use Scalar::Util;
+
 require Exporter;
 
 use vars qw ($VERSION @ISA @EXPORT);
 
-$VERSION = '1.00';
+$VERSION = '2.00';
 @ISA = qw (Exporter);
 @EXPORT = qw (walk walkdepth);
 
-use vars qw ($container $type $seen $address $depth);
+use vars qw ($container $type $seen $address $depth $index $key);
 
 # Forward declarations.
 sub walk;
@@ -45,8 +45,8 @@ sub __recurse;
 sub walk {
     my ($options, @args) = @_;
    
-    unless ('HASH' eq ref $options) {
-	$options = { wanted => $options };
+    unless (UNIVERSAL::isa($options, 'HASH')) {
+        $options = { wanted => $options };
     }
 
     __walk ($options, @args);
@@ -55,8 +55,8 @@ sub walk {
 sub walkdepth {
     my ($options, @args) = @_;
 
-    unless ('HASH' eq ref $options) {
-	$options = { wanted => $options };
+    unless (UNIVERSAL::isa($options, 'HASH')) {
+        $options = { wanted => $options };
     }
 
     $options->{bydepth} = 1;
@@ -68,12 +68,28 @@ sub __walk {
     my ($options, @args) = @_;
 
     $options->{seen} = {};
-    $options->{copy} = 1 unless exists $options->{copy};
 
+    local $index = 0;
     foreach my $item (@args) {
-    	local $depth;
-	$depth = 0;
-	__recurse $options, $item;
+        local ($container, $type, $depth);
+        if (ref $item) {
+            if (UNIVERSAL::isa ($item, 'HASH')) {
+                $container = $item;
+                $type = 'HASH';
+            } elsif (UNIVERSAL::isa ($item, 'ARRAY')) {
+                $container = $item;
+                $type = 'ARRAY';
+            } else {
+                $container = \@args;
+                $type = 'ARRAY';
+            }
+        } else {
+            $container = \@args;
+            $type = 'ARRAY';
+        }
+        $depth = 0;
+        __recurse $options, $item;
+        ++$index;
     }
     
     return 1;
@@ -85,81 +101,88 @@ sub __recurse {
     ++$depth;
     
     my @children;
-    my $data_type;
+    my $data_type = '';
 
-    local ($address, $seen);
-    undef $address;
-    $seen = 0;
+    local ($container, $type, $address, $seen) = ($container, $type, undef, 0);
     my $ref = ref $item;
 
     if ($ref) {
-	my $blessed = -1 != index $ref, '=';
+        my $blessed = Scalar::Util::blessed($item);
 
-	# Avoid fancy overloading stuff.
-	bless $item if $blessed;
-	$address = int $item;
-	
-	$seen = $options->{seen}->{$address}++;
+        # Avoid fancy overloading stuff.
+        bless $item if $blessed;
+        $address = Scalar::Util::refaddr($item);
+    
+        $seen = $options->{seen}->{$address}++;
 
-	if (UNIVERSAL::isa ($item, 'HASH')) {
-		$data_type = 'HASH';
-	} elsif (UNIVERSAL::isa ($item, 'ARRAY')) {
-		$data_type = 'ARRAY';
-	} else {
-		$data_type = '';
-	}
-	
-	if ($data_type eq 'HASH' || $data_type eq 'ARRAY') {
-	    if (('ARRAY' eq $data_type || 'HASH' eq $data_type)) {
-		if ('ARRAY' eq $data_type) {
-		    @children = @{$item};
-		} else {
-		    @children = %{$item};
-		}
-		
-		if ($options->{copy}) {
-		    if ('ARRAY' eq $data_type) {
-			@children = $options->{preprocess} (@{$item}) 
-			    if $options->{preprocess};
-		    } else {
-			@children = %{$item};
-			@children = $options->{preprocess} (@children) 
-			    if $options->{preprocess};
-			@children = $options->{preprocess_hash} (@children) 
-			    if $options->{preprocess_hash};
-		    }
-		} else {
-		    $item = $options->{preprocess} ($item) 
-			if $options->{preprocess};
-		    $item = $options->{preprocess_hash} ($item) 
-			if 'HASH' eq $data_type && $options->{preprocess_hash};
-		    @children = 'HASH' eq $data_type ? %{$item} : @{$item};
-		}
-	    }
-	}
+        if (UNIVERSAL::isa ($item, 'HASH')) {
+            $data_type = 'HASH';
+        } elsif (UNIVERSAL::isa ($item, 'ARRAY')) {
+            $data_type = 'ARRAY';
+        } else {
+            $data_type = '';
+        }
+    
+        if ('ARRAY' eq $data_type || 'HASH' eq $data_type) {
+            local $index = -1;
+            local $type = $data_type;
+            local $container = $item;
+
+            if ('ARRAY' eq $data_type) {
+                @children = @{$item};
+            } else {
+                @children = %{$item};
+            }
+        
+            if ('ARRAY' eq $data_type) {
+                @children = $options->{preprocess} (@{$item}) 
+                        if $options->{preprocess};
+            } else {
+                local $container = \@children;
+                @children = $options->{preprocess} (@children) 
+                        if $options->{preprocess};
+                @children = $options->{preprocess_hash} (@children) 
+                        if $options->{preprocess_hash};
+            }
+        } else {
+            $data_type = '';
+        }
+
+        # Recover original object state.
+        bless $item, $ref if $blessed;
     }
 
     unless ($options->{bydepth}) {
-	$_ = $item;
-	$options->{wanted}->($item);
+        local $_ = $item;
+        $options->{wanted}->($item);
     }
 
-    local ($container, $type);
-    $type = $data_type;
-    $container = $item;
+    if (@children && ($options->{follow} || !$seen)) {
+        local ($container, $type, $index);
+        $type = $data_type;
+        $container = $item;
+        $index = 0;
 
-    if ($options->{follow} || !$seen) {
-	foreach my $child (@children) {
-	    __recurse $options, $child;
-	}
+        foreach my $child (@children) {
+            if ($type eq 'HASH' && $index & 1) {
+                $key = $children[$index - 1];
+            } else {
+                undef $key;
+            }
+            __recurse $options, $child;
+            ++$index;
+        }
     }
 
     if ($options->{bydepth}) {
-	$_ = $item;
-	$options->{wanted}->($item);
+        local $_ = $item;
+        $options->{wanted}->($item);
     }
 
-    $options->{postprocess}->() if $options->{postprocess};
+    if ($data_type) {
+        local ($container, $type, $index) = ($item, $data_type, -1);
+        $options->{postprocess}->() if $options->{postprocess};
+    }
 
     --$depth;
     # void
@@ -254,10 +277,9 @@ preprocessing function is called before the loop that calls the
 C<wanted()> function.  It is called with a list of member nodes
 and is expected to return such a list.  The list will contain
 all sub-nodes, regardless of the value of the option I<follow>!
-The list is normally a shallow copy of the data contained in the original
+The list is a shallow copy of the data contained in the original
 structure.  You can therefore safely delete items in it, without
-affecting the original data.  You can use the option I<copy>,
-if you want to change that behavior.
+affecting the original data.
 
 The behavior is identical for regular arrays and hashes, so you 
 probably want to coerce the list passed as an argument into a hash 
@@ -297,29 +319,6 @@ Please note that the &wanted function is also called for nodes
 that have already been visited!  The effect of I<follow> is to
 suppress descending into subnodes.  
 
-=item B<copy>
-
-Normally, the &preprocess function is called with a shallow copy
-of the data.  If you set the option I<copy> to a false value,
-the &preprocess function is called with one single argument,
-a reference to the original data structure.  In that case, you
-also have to return a suitable reference.
-
-Using this option will result in a slight performance win, and
-can make it sometimes easier to manipulate the original data.
-
-What is a shallow copy?  Think of a list containing references
-to hashes:
-
-    my @list = ({ foo => 'bar' }, { foo => 'baz' });
-    my @shallow = @list;
-
-After this, @shallow will contain a new list, but the items
-stored in it are exactly identical to the ones stored in the
-original.  In other words, @shallow occupies new memory, whereas
-both lists contain references to the same memory for the list
-members.
-
 =back
 
 All other options are silently ignored.
@@ -349,7 +348,8 @@ a hash or an array.  Think "directory" in terms of File::Find(3pm)!
 =item B<$Data::Walk::type>
 
 The base type of the object that $Data::Walk::container
-references.  This is either "ARRAY" or "HASH".
+references.  This is either "ARRAY" or "HASH" or the empty string for
+everything else.
 
 =item B<$Data::Walk::seen>
 
@@ -368,6 +368,18 @@ references, the value is undefined.
 
 The depth of the current recursion.
 
+=item B<$Data::Walk::index>
+
+Holds the index of the current item in the container.  Note that hashes
+and arrays are treated the same.  Therefore, if the current container is
+a hash and B<$Data::Walk::index> is even then B<$_> is a hash key.  If
+it is odd, then B<$_> is a hash value.
+
+Note that the root container is the array of items to search that you
+passed to the wanted function!
+
+This variable has been added in Data::Walk version 1.01.
+
 =back
 
 These variables should not be modified.
@@ -418,25 +430,10 @@ I<follow_skip>, I<no_chdir>, I<untaint>, I<untaint_pattern>, and
 I<untaint_skip>.  To give truth the honor, all unrecognized options
 are skipped.
 
-You may argue, that the options I<untaint> and friends would be
-useful, too, allowing you to recursively untaint data structures.
-But, hey, that is what Data::Walk(3pm) is all about.  It makes
-it very easy for you to write that yourself.
-
 =head1 EXAMPLES
 
 Following are some recipies for common tasks.  
 
-=head2 Recursive Untainting
-
-    sub untaint { 
-    	s/(.*)/$1/s unless ref $_;
-    };
-    walk \&untaint, $data;
-
-See perlsec(1), if you don't understand why the untaint() function
-untaints your data here.
-
 =head2 Recurse To Maximum Depth
 
 If you want to stop the recursion at a certain level, do it as follows:
@@ -444,13 +441,13 @@ If you want to stop the recursion at a certain level, do it as follows:
     my $max_depth = 20;
     sub not_too_deep {
         if ($Data::Walk::depth > $max_depth) {
-	    return ();
+        return ();
         } else {
-	    return @_;
+        return @_;
         }
     }
     sub do_something1 {
-    	# Your code goes here.
+        # Your code goes here.
     }
     walk { wanted => \&do_something, preprocess => \&not_too_deep };
 
@@ -461,7 +458,7 @@ bug tracking system at http://rt.cpan.org/NoAuth/Bugs.html?Dist=Data-Walk.
 
 =head1 COPYING
 
-Copyright (C) 2005-2006, Guido Flohr E<lt>guido@imperia.netE<gt>, all
+Copyright (C) 2005-2006, Guido Flohr E<lt>guido.flohr@cantanea.comE<gt>, all
 rights reserved.
 
 This program is free software; you can redistribute it and/or modify it
